/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.servicemix.cxfbc;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;

import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.apache.cxf.Bus;
import org.apache.cxf.binding.AbstractBindingFactory;
import org.apache.cxf.binding.soap.Soap11;
import org.apache.cxf.binding.soap.Soap12;
import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.SoapHeaderOutFilterInterceptor;
import org.apache.cxf.binding.soap.interceptor.SoapOutInterceptor;
import org.apache.cxf.binding.soap.interceptor.SoapPreProtocolOutInterceptor;
import org.apache.cxf.bus.spring.SpringBusFactory;
import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.configuration.security.FiltersType;
import org.apache.cxf.endpoint.Endpoint;
import org.apache.cxf.endpoint.EndpointImpl;
import org.apache.cxf.interceptor.AttachmentOutInterceptor;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.interceptor.StaxOutInterceptor;
import org.apache.cxf.message.Exchange;
import org.apache.cxf.message.ExchangeImpl;
import org.apache.cxf.message.Message;
import org.apache.cxf.message.MessageImpl;
import org.apache.cxf.phase.PhaseChainCache;
import org.apache.cxf.phase.PhaseInterceptorChain;
import org.apache.cxf.phase.PhaseManager;
import org.apache.cxf.service.Service;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.EndpointInfo;
import org.apache.cxf.service.model.ServiceInfo;
import org.apache.cxf.transport.Conduit;
import org.apache.cxf.transport.ConduitInitiator;
import org.apache.cxf.transport.ConduitInitiatorManager;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transport.jbi.JBIMessageHelper;
import org.apache.cxf.wsdl.WSDLConstants;
import org.apache.cxf.wsdl11.WSDLServiceFactory;
import org.apache.servicemix.common.endpoints.ProviderEndpoint;
import org.apache.servicemix.cxfbc.interceptors.JbiOutInterceptor;
import org.apache.servicemix.cxfbc.interceptors.JbiOutWsdl1Interceptor;
import org.apache.servicemix.cxfbc.interceptors.MtomCheckInterceptor;
import org.apache.servicemix.soap.util.DomUtil;
import org.dom4j.Document;
import org.dom4j.Node;
import org.dom4j.XPath;
import org.dom4j.io.SAXReader;
import org.dom4j.xpath.DefaultXPath;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.Resource;

import com.ibm.wsdl.Constants;



/**
 * 
 * @author gnodet
 * @org.apache.xbean.XBean element="provider"
 */
public class CxfBcProvider extends ProviderEndpoint implements
        CxfBcEndpointWithInterceptor {

    private static final Logger LOG = LogUtils.getL7dLogger(org.apache.servicemix.cxfbc.CxfBcProvider.class);
    
    
    List<Interceptor> in = new CopyOnWriteArrayList<Interceptor>();

    List<Interceptor> out = new CopyOnWriteArrayList<Interceptor>();

    List<Interceptor> outFault = new CopyOnWriteArrayList<Interceptor>();

    List<Interceptor> inFault = new CopyOnWriteArrayList<Interceptor>();
    
    private Resource wsdl;
    
    private String busCfg;
    
    private Bus bus;
    
    private URI locationURI;
    
    private EndpointInfo ei;
    
    private Endpoint ep;

    private Conduit conduit;
    
    private boolean mtomEnabled;
    
    public void processExchange(MessageExchange exchange) {
        
    }

    public void process(MessageExchange exchange) throws Exception {
        // Skip done exchanges
        if (exchange.getStatus() == ExchangeStatus.DONE) {
            return;
        }

        NormalizedMessage nm = exchange.getMessage("in");
                       
        CxfBcProviderMessageObserver obs = new CxfBcProviderMessageObserver(exchange, this);
        conduit.setMessageObserver(obs);
        SoapMessage message = new SoapMessage(new MessageImpl());
        message.put(MessageExchange.class, exchange);
        Exchange cxfExchange = new ExchangeImpl();
        message.setExchange(cxfExchange);
        
        /////////// Serve per far funzionare WS-Adressing
        cxfExchange.setOutMessage(message);
        ///////////
        
        boolean mtomEnabled = isMtomEnabled(nm);
        QName opeName = exchange.getOperation();
        BindingOperationInfo boi = null;
        if (opeName == null) {
            // if interface only have one operation, may not specify the opeName in MessageExchange
            boi = ei.getBinding().getOperations().iterator().next();
        } else {
            boi = ei.getBinding().getOperation(exchange.getOperation());   
        }
         
        cxfExchange.put(BindingOperationInfo.class, boi);
        cxfExchange.put(Endpoint.class, ep);
        
        // SOAP 1.2 support
        String bindingId = boi.getBinding().getBindingId();
        if (WSDLConstants.SOAP11_NAMESPACE.equals(bindingId)) {
            message.setVersion(Soap11.getInstance());
        } else if (WSDLConstants.SOAP12_NAMESPACE.equals(bindingId)) {
            message.setVersion(Soap12.getInstance());
        }
	    ///////////
        
        PhaseChainCache outboundChainCache = new PhaseChainCache();
        PhaseManager pm = getBus().getExtension(PhaseManager.class);
        List<Interceptor> outList = new ArrayList<Interceptor>();
        outList.add(new JbiOutInterceptor(false));
        outList.add(new MtomCheckInterceptor(mtomEnabled));
        outList.add(new AttachmentOutInterceptor());
        outList.add(new JbiOutWsdl1Interceptor());
        outList.add(new SoapPreProtocolOutInterceptor());
        outList.add(new SoapHeaderOutFilterInterceptor());
        outList.add(new SoapOutInterceptor(getBus()));
                
        
        PhaseInterceptorChain outChain = outboundChainCache.get(pm.getOutPhases(), outList);
        outChain.add(getOutInterceptors());
        outChain.add(getOutFaultInterceptors());


        message.setInterceptorChain(outChain);
        
        InputStream is = JBIMessageHelper.convertMessageToInputStream(nm.getContent());
        
        StreamSource source = new StreamSource(is);
        message.setContent(Source.class, source);
        message.setContent(InputStream.class, is);
        

        // We MUST set the content type, otherwise the AttachmentSerializer used in AttachmentOutInterceptor
        // throws an Exception
        message.put(Message.CONTENT_TYPE, "application/soap+xml");
        
        conduit.prepare(message);
        OutputStream os = message.getContent(OutputStream.class);
        
        XMLStreamWriter writer = message.getContent(XMLStreamWriter.class);
        

        String encoding = getEncoding(message);
        
        try {
            writer = StaxOutInterceptor.getXMLOutputFactory(message).createXMLStreamWriter(os, encoding);
        } catch (XMLStreamException e) {
            //
        }
        message.setContent(XMLStreamWriter.class, writer);
        message.put(org.apache.cxf.message.Message.REQUESTOR_ROLE, true);
        outChain.doIntercept(message);
        XMLStreamWriter xtw = message.getContent(XMLStreamWriter.class);
        if (xtw != null) {
            xtw.writeEndDocument();
            xtw.close();
        }
        
        os.flush();
        is.close();
        os.close();
    }
    
    public List<Interceptor> getOutFaultInterceptors() {
        return outFault;
    }

    public List<Interceptor> getInFaultInterceptors() {
        return inFault;
    }

    public List<Interceptor> getInInterceptors() {
        return in;
    }

    public List<Interceptor> getOutInterceptors() {
        return out;
    }

    public void setInInterceptors(List<Interceptor> interceptors) {
        in = interceptors;
    }

    public void setInFaultInterceptors(List<Interceptor> interceptors) {
        inFault = interceptors;
    }

    public void setOutInterceptors(List<Interceptor> interceptors) {
        out = interceptors;
    }

    public void setOutFaultInterceptors(List<Interceptor> interceptors) {
        outFault = interceptors;
    }

    public void setWsdl(Resource wsdl) {
        this.wsdl = wsdl;
    }

    public Resource getWsdl() {
        return wsdl;
    }
    
    @Override
    public void validate() throws DeploymentException {
        try {
            if (definition == null) {
                if (wsdl == null) {
                    throw new DeploymentException("wsdl property must be set");
                }
                description = DomUtil.parse(wsdl.getInputStream());
                WSDLFactory wsdlFactory = WSDLFactory.newInstance();
                WSDLReader reader = wsdlFactory.newWSDLReader();
                reader.setFeature(Constants.FEATURE_VERBOSE, false);
                definition = reader.readWSDL(wsdl.getURL().toString(),
                        description);
                
                
                
                WSDLServiceFactory factory = new WSDLServiceFactory(getBus(),
                        definition, service);
                Service cxfService = factory.create();
                
                ei = cxfService.getServiceInfos().iterator().next()
                        .getEndpoints().iterator().next();
                for (ServiceInfo serviceInfo : cxfService.getServiceInfos()) {
                    if (serviceInfo.getName().equals(service)
                        && getEndpoint() != null 
                        && serviceInfo.getEndpoint(new QName(
                                serviceInfo.getName().getNamespaceURI(), getEndpoint())) != null) {
                        ei = serviceInfo.getEndpoint(new QName(
                                    serviceInfo.getName().getNamespaceURI(), getEndpoint()));
                 
                    }
                }

                if (endpoint == null) {
                    endpoint = ei.getName().getLocalPart();
                }
                ei.getBinding().setProperty(
                        AbstractBindingFactory.DATABINDING_DISABLED, Boolean.TRUE);
                
                if (locationURI == null) {
                    // if not specify target address, get it from the wsdl
                    locationURI = new URI(ei.getAddress());
                    LOG.fine("address is " + locationURI.toString());
                }
                ep = new EndpointImpl(getBus(), cxfService, ei);
                
                //init transport
                ei.setAddress(locationURI.toString());
                ConduitInitiatorManager conduitMgr = getBus().getExtension(ConduitInitiatorManager.class);
                ConduitInitiator conduitInit = conduitMgr.getConduitInitiator("http://schemas.xmlsoap.org/soap/http");

                conduit = conduitInit.getConduit(ei);
                if (conduit instanceof HTTPConduit) {
                	if (locationURI.toString().startsWith("https")){
                        
                		((HTTPConduit)conduit).setTlsClientParameters(getTLSClientParametersFromBus());
                	}
					
				}
                
              
                

                super.validate();
            }
        } catch (DeploymentException e) {
            throw e;
        } catch (Exception e) {
            throw new DeploymentException(e);
        }
    }

    @Override
    public void start() throws Exception {
        super.start();
        
    }
    
    protected Bus getBus() {
        if (getBusCfg() != null) {
            if (bus == null) {
                SpringBusFactory bf = new SpringBusFactory();
                bus = bf.createBus(getBusCfg());
            }
            return bus;
        } else {
            return ((CxfBcComponent) getServiceUnit().getComponent()).getBus();
        }
    }
    
    public void setBusCfg(String busCfg) {
        this.busCfg = busCfg;
    }

    public String getBusCfg() {
        return busCfg;
    }

    public void setLocationURI(URI locationURI) {
        this.locationURI = locationURI;
    }

    public URI getLocationURI() {
        return locationURI;
    }
    
    private String getEncoding(Message message) {
        Exchange ex = message.getExchange();
        String encoding = (String)message.get(Message.ENCODING);
        if (encoding == null && ex.getInMessage() != null) {
            encoding = (String) ex.getInMessage().get(Message.ENCODING);
            message.put(Message.ENCODING, encoding);
        }
        
        if (encoding == null) {
            encoding = "UTF-8";
            message.put(Message.ENCODING, encoding);
        }
        return encoding;
    }
    
    Endpoint getCxfEndpoint() {
        return this.ep;
    }
    
    EndpointInfo getEndpointInfo() {
        return this.ei;
    }

    public void setMtomEnabled(boolean mtomEnabled) {
        this.mtomEnabled = mtomEnabled;
    }

    public boolean isMtomEnabled(NormalizedMessage nm) {
    	// Dont use MTOM if there are no attachments in the Normalized message
        return mtomEnabled && (nm.getAttachmentNames() != null) && (nm.getAttachmentNames().size() > 0);
    }
    

    

	public static KeyManager[] getKeyManagers(KeyStore keyStore,
			String keyPassword) throws GeneralSecurityException, IOException {
		// For tests, we just use the default algorithm
		String alg = KeyManagerFactory.getDefaultAlgorithm();

		char[] keyPass = keyPassword != null ? keyPassword.toCharArray() : null;

		// For tests, we just use the default provider.
		KeyManagerFactory fac = KeyManagerFactory.getInstance(alg);

		fac.init(keyStore, keyPass);

		return fac.getKeyManagers();
	}

	public static TrustManager[] getTrustManagers(KeyStore keyStore)
			throws GeneralSecurityException, IOException {
		// For tests, we just use the default algorithm
		String alg = TrustManagerFactory.getDefaultAlgorithm();

		// For tests, we just use the default provider.
		TrustManagerFactory fac = TrustManagerFactory.getInstance(alg);

		fac.init(keyStore);

		return fac.getTrustManagers();
	}

	public static KeyStore getKeyStore(String ksType, String file,
			String ksPassword) throws GeneralSecurityException, IOException {

		String type = ksType != null ? ksType : KeyStore.getDefaultType();

		char[] password = ksPassword != null ? ksPassword.toCharArray() : null;

		// We just use the default Keystore provider
		KeyStore keyStore = KeyStore.getInstance(type);
		
		InputStream is = new ClassPathResource(file).getInputStream();
		keyStore.load(is, password);

		return keyStore;
	}
	
	public TLSClientParameters getTLSClientParametersFromBus() throws Exception {
		TLSClientParameters tlsParams = new TLSClientParameters();
		tlsParams.setSecureSocketProtocol("SSL");
		
		InputStream busCfgIn = new ClassPathResource(getBusCfg()).getInputStream();
		SAXReader saxReader = new SAXReader();
		
		
		Document busDocument = saxReader.read(busCfgIn);
	
		
		XPath xPathExpression = new DefaultXPath("/beans/http:conduit/http:tlsClientParameters/sec:keyManagers");
		xPathExpression.setNamespaceURIs(getNamespaceURI());
		Node keyManagerNode = (Node)xPathExpression.evaluate(busDocument);
		
		xPathExpression = new DefaultXPath("/beans/http:conduit/http:tlsClientParameters/sec:keyManagers/sec:keyStore");
		xPathExpression.setNamespaceURIs(getNamespaceURI());
		Node keyStoreNode = (Node)xPathExpression.evaluate(busDocument);
		
		xPathExpression = new DefaultXPath("/beans/http:conduit/http:tlsClientParameters/sec:trustManagers/sec:keyStore");
		xPathExpression.setNamespaceURIs(getNamespaceURI());
		Node trustStoreNode = (Node)xPathExpression.evaluate(busDocument);
		
		
	
		
		String keyManagerPassword = keyManagerNode.valueOf("@keyPassword");
		String keyStoreType= keyStoreNode.valueOf("@type");
		String keyStorePassword= keyStoreNode.valueOf("@password");
		String keyStoreResource= keyStoreNode.valueOf("@resource");
		
		String trustStoreType= trustStoreNode.valueOf("@type");
		String trustStorePassword= trustStoreNode.valueOf("@password");
		String trustStoreResource= trustStoreNode.valueOf("@resource");
		
		

						
		
		
		
        KeyManager[] kmgrs = getKeyManagers(getKeyStore(keyStoreType, keyStoreResource, keyStorePassword), keyManagerPassword);
        TrustManager[] tmgrs = getTrustManagers(getKeyStore(trustStoreType, trustStoreResource, trustStorePassword));
        
        tlsParams.setKeyManagers(kmgrs);
        tlsParams.setTrustManagers(tmgrs);
        
       
        FiltersType filters = new FiltersType();
        
        filters.getInclude().add(".*_EXPORT_.*");
        filters.getInclude().add(".*_EXPORT1024_.*");
        filters.getInclude().add(".*_WITH_DES_.*");
        filters.getInclude().add(".*_WITH_NULL_.*");
        filters.getInclude().add(".*_DH_anon_.*");
        
        filters.getInclude().add(".*_WITH_AES_.*");
        filters.getInclude().add(".*_WITH_3DES_.*");
        
        tlsParams.setCipherSuitesFilter(filters);
        return tlsParams;
	}
	
	public Map<String, String> getNamespaceURI(){
		Map<String, String> nsMap = new HashMap<String, String>();
		nsMap.put("sec","http://cxf.apache.org/configuration/security");
		nsMap.put("http","http://cxf.apache.org/transports/http/configuration");
		nsMap.put("httpj","http://cxf.apache.org/transports/http-jetty/configuration");
		nsMap.put("jaxws","http://java.sun.com/xml/ns/jaxws");
		return nsMap;
		
	}
	
}
